// SPDX-License-Identifier: GPL-2.0
/*
 *  Copyright 2022 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <asm/cache.h>
#include <asm/global_data.h>
#include <asm/unaligned.h>
#include <bootm.h>
#include <command.h>
#include <common.h>
#include <fs.h>
#include <image.h>
#include <linux/sizes.h>
#include <log.h>
#include <malloc.h>
#include <part.h>
#include <u-boot/crc.h>

#include "ssbootmem.h"

DECLARE_GLOBAL_DATA_PTR;

/* Decompress 'n' lz4 blocks from a 'compressed' buffer and put the result in an 'elf' buffer */
static int decompress_lz4(int n, struct ss_ctx *ctx)
{
	int ret;
	void *in   = buf_cur(&ctx->cbuf);
	void *iend = ctx->cbuf.data + ctx->cbuf.size;
	void *out  = buf_cur(&ctx->ebuf);
	void *oend = ctx->ebuf.data + ctx->ebuf.size;

	while (n--) {
		u32 block_header, block_size;

		block_header = get_unaligned_le32(in);
		in += sizeof(u32);
		block_size = block_header & ~LZ4F_BLOCKUNCOMPRESSED_FLAG;

		if (in + block_size > iend) {
			ret = -EINVAL;
			break;
		}
		if (!block_size) {
			ret = 0;
			break;
		}
		if (block_header & LZ4F_BLOCKUNCOMPRESSED_FLAG) {
			if (out + block_size > oend) {
				ret = -ENOBUFS;
				break;
			} else {
				memcpy(out, in, block_size);
				out += block_size;
				ret = block_size;
			}
		} else {
			/*
			 * copy the lz4 block to heap and decompress it
			 * directly into the destination elf file address
			 */
			memcpy(ctx->tbuf.data, in, block_size);
			ret = LZ4_decompress_generic(ctx->tbuf.data, out, block_size,
					oend - out, endOnInputSize, full, 0,
					noDict, out, NULL, 0);
			if (ret < 0) {
				ret = -EPROTO;
				break;
			}
			out += ret;
			debug("lz4 block decoded: size %d B -> %d B\n",
					block_size, ret);
		}
		in += block_size;
		if (ctx->has_block_checksum)
			in += sizeof(u32);
	}
	ctx->cbuf.start += in - buf_cur(&ctx->cbuf);
	ctx->ebuf.start += out - buf_cur(&ctx->ebuf);

	return ret;
}

#ifdef CONFIG_CMD_SSBOOT_CKSUM
static unsigned int ssboot_calc_data_checksum(Elf_Ehdr *ehdr, Elf_Phdr *phdr)
{
	int i;
	unsigned int data_cksum = 0;

	for (i = 0; i < ehdr->e_phnum; i++, phdr++) {
		if (phdr->p_type != PT_LOAD)
			continue;

		if (phdr->p_filesz == 0)
			continue;

		data_cksum = crc32(data_cksum, (void *)phdr->p_paddr,
				   phdr->p_filesz);
#ifdef CONFIG_CMD_SSBOOT_DEBUG_CORRUPTION
		printf("check %p -> %p: %08x\n", (void *) phdr->p_paddr,
				(void *) (phdr->p_paddr + phdr->p_filesz),
				data_cksum);
#endif
	}

	return data_cksum;
}

/* Check the integrity of loaded ELF segments using crc32 checksum */
static int ssboot_verify_checksum(Elf_Ehdr *ehdr, Elf_Phdr *phdr)
{
	int i;
	Elf_Phdr *cksum_phdr = NULL;
	unsigned int ehdr_cksum, phdr_cksum, data_cksum;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
	printf("## Verifying snapshot checksums...\n");

	for (i = 0; i < ehdr->e_phnum; i++) {
		if (phdr[i].p_type != PT_LOAD) {
			if (phdr[i].p_offset == CKSUM_TYPE_CRC32)
				cksum_phdr = &phdr[i];
			continue;
		}
		break;
	}

	if (cksum_phdr == NULL) {
		printf("Warning: checksum not found in image, skipping verification\n");
		return 0;
	}

	ehdr_cksum = crc32(0, (void *)ehdr, sizeof(*ehdr));
	phdr_cksum = crc32(0, (void *)&cksum_phdr[1], ehdr->e_phnum * ehdr->e_phentsize
			- ((size_t)&cksum_phdr[1] - (size_t)phdr));
	data_cksum = ssboot_calc_data_checksum(ehdr, phdr);

	if (elf_ehdr_cksum(cksum_phdr) != ehdr_cksum
		|| elf_phdr_cksum(cksum_phdr) != phdr_cksum
		|| elf_data_cksum(cksum_phdr) != data_cksum) {
		printf("ehdr_cksum: %08x %08x\n",
			(unsigned int) elf_ehdr_cksum(cksum_phdr), ehdr_cksum);
		printf("phdr_cksum: %08x %08x\n",
			(unsigned int) elf_phdr_cksum(cksum_phdr), phdr_cksum);
		printf("data_cksum: %08x %08x\n",
			(unsigned int) elf_data_cksum(cksum_phdr), data_cksum);

		return -1;
	}

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("\tVerification time: %lu ms\n", time);
#endif
	return 0;
}
#endif /* CONFIG_CMD_SSBOOT_CKSUM */

/* Extract the ELF header and program header information from an ELF file in memory */
static int get_elf_headers(struct ss_elf *elf, void *addr)
{
	loff_t phdrs_size;

	memcpy(&elf->ehdr, addr, sizeof(Elf_Ehdr));
	phdrs_size = elf->ehdr.e_phnum * elf->ehdr.e_phentsize;

	elf->phdrs = malloc(phdrs_size);
	if (!elf->phdrs) {
		printf("Error: cannot allocate %llu B memory for program headers!\n",
				phdrs_size);
		return -1;
	}

	memcpy(elf->phdrs, addr + elf->ehdr.e_phoff, phdrs_size);

	return 0;
}

/*
 * The LZ4 format does not specify storing the total compressed data size in the header.
 * Calculate the size explicitly by summing up individual blocks sizes.
 */
static loff_t get_lz4_file_size(void *in, struct ss_ctx *ctx)
{
	u32 block_header, block_size;
	void *cur = in;

	while (1) {
		block_header = get_unaligned_le32(cur);
		block_size = block_header & ~LZ4F_BLOCKUNCOMPRESSED_FLAG;

		if (block_size > ctx->max_block_size) {
			printf("Error: lz4 block size exceeds the max block size!\n");
			return 0;
		}
		if (block_size == 0) {
			cur += sizeof(u32);
			if (ctx->has_content_checksum)
				cur += sizeof(u32);
			break;
		}
		cur += sizeof(u32) + block_size
			+ (ctx->has_block_checksum ? sizeof(u32) : 0);
	}
	return cur - in;
}

/* Determine the end of unused u-boot memory which is treated as a free buffer */
static ulong get_end_of_mem(ulong *free_size_ret)
{
	int i, iend;
	struct lmb lmb;
	ulong memstart, memend;
	ulong free_start, free_size;
	ulong lres_base, lres_size;

	if (!gd->fdt_blob) {
		printf("Error: fdt_blob not found!\n");
		return 0;
	}

	memstart = gd->ram_base;
	memend = gd->ram_top;
	lres_base = memstart;

	lmb_init_and_reserve(&lmb, gd->bd, (void *)gd->fdt_blob);
	lmb_dump_all(&lmb);

	/* get the last reserved region by address */
	for (i = 0, iend = 0; i < lmb.reserved.cnt; i++) {
		if (lmb.reserved.region[i].base > lres_base) {
			lres_base = lmb.reserved.region[i].base;
			lres_size = lmb.reserved.region[i].size;
			iend = i;
		}
	}

	/* the last lmb reserved region must be at the end of memory */
	if (lres_base + lres_size != memend) {
		printf("Warning: the last reserved region (0x%08lx-0x%08lx)\n"
				"\tnot at the end of memory (0x%08lx)\n",
				lres_base, lres_base + lres_size, memend);
	}

	if (iend) {
		free_start = (ulong)lmb.reserved.region[iend - 1].base
				+ (ulong)lmb.reserved.region[iend - 1].size;
		free_size = (ulong)lmb.reserved.region[iend].base - free_start;
	} else {
		free_start = memstart;
		free_size = (ulong)lmb.reserved.region[iend].base - free_start;
	}

	printf("Using memory region 0x%08lx-0x%08lx for in-place loading\n", free_start,
			free_start + free_size);

	*free_size_ret = free_size;
	return free_start + free_size;
}

static int have_enough_memory(struct ss_ctx *ctx, loff_t size)
{
	if (!ctx->memend)
		ctx->memend = get_end_of_mem(&ctx->free_size);

	if (ctx->memend < size) {
		printf("Error: snapshot image size (0x%08llx) exceeds the memory space\n"
				"\tavailable (0x%08lx) for in-place loading!\n"
				"Please check your reserved memory configuration.\n",
				size, ctx->free_size);
		return -1;
	}

	return 0;
}

/* Move the ELF segments in the snapshot image to their actual memory addresses */
static int load_elf_from_mem(struct ss_ctx *ctx, void *addr)
{
	int i, ret;
	loff_t elf_file_size;
	Elf_Ehdr *ehdr;
	Elf_Phdr *phdrs;
	Elf_Phdr *cur, *prev;
	ulong memstart, filestart;
	ulong lastseg_addr, lastseg_size;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
	printf("## Restoring ELF snapshot from address 0x%08lx...\n", (ulong)addr);

	ret = get_elf_headers(&ctx->elf, addr);
	if (ret < 0)
		return -1;

	ehdr = &ctx->elf.ehdr;
	phdrs = ctx->elf.phdrs;

	elf_file_size = phdrs[ehdr->e_phnum - 1].p_filesz
			+ phdrs[ehdr->e_phnum - 1].p_offset;

	if (ctx->total_read_len && elf_file_size != ctx->total_read_len) {
		printf("Error: elf file truncated! Expected %d B, got %d B\n",
			(unsigned) ctx->total_read_len, (unsigned) elf_file_size);
		return -1;
	}

	if (ctx->image_size_elf && elf_file_size != ctx->image_size_elf) {
		printf("Error: elf file size calculated wrong!\n");
		return -1;
	}
	ctx->image_size_elf = elf_file_size;

	if (have_enough_memory(ctx, elf_file_size) < 0)
		return -1;

	/*
	 * check if the memory layout seen by uboot is the same as seen by kernel
	 * when the snapshot was created. we need to able to load the last elf
	 * segment within the end-of-memory boundary.
	 */
	lastseg_addr = phdrs[ehdr->e_phnum - 1].p_paddr;
	lastseg_size = phdrs[ehdr->e_phnum - 1].p_memsz;
	if (ctx->memend < lastseg_addr + lastseg_size) {
		struct lmb lmb;
		ulong pmmem_base, pmmem_size;

		lmb_init_and_reserve(&lmb, gd->bd, (void *)gd->fdt_blob);

		printf("Error: elf segment(s) spill over the determined end-of-memory address!\n"
			"The last segment needs to be loaded at the address 0x%08lx-0x%08lx\n"
			"but the end-of-memory was calcluated to be 0x%08lx.\n",
			lastseg_addr, lastseg_addr + lastseg_size,
			ctx->memend);
		lmb_dump_all_force(&lmb);
		pmmem_base = env_get_hex("uboot_pmmem_base", ~0);
		pmmem_size = env_get_hex("uboot_pmmem_size", ~0);
		printf("PMMEM reserved region: 0x%08lx-0x%08lx\n\n", pmmem_base,
				pmmem_base + pmmem_size);
		if (ctx->memend != pmmem_base) {
			printf("The boundary address PMMEM passed onto kernel is different\n"
				"from what was reserved in the LMB.\n");
		} else {
			printf("Since the PMMEM base address and the end-of-memory addresses\n"
				"are same, it is likely that the PMMEM region is not getting\n"
				"reserved properly in the kernel! Check the regions in:\n"
				"/sys/kernel/debug/memblock/reserved\n");
		}
		return -1;
	}

	memstart = ctx->memend - elf_file_size;

	/* shift the elf file data to the end of usable memory */
	if ((ulong)addr < memstart)
		memmove((void *)memstart, addr, elf_file_size);
	else
		memstart = (ulong)addr;

	filestart = memstart;

	printf("\tELF snapshot image size:  %llu.%03llu MB\n",
			ctx->image_size_elf >> 20, FLOAT(ctx->image_size_elf, 10));

	prev = phdrs;
	for (i = 0; i < ehdr->e_phnum; i++) {
		cur = phdrs + i;

#ifdef CONFIG_CMD_SSBOOT_CMDLINEINFO
		if (cur->p_type == PT_NOTE) {
			if (cur->p_paddr < gd->ram_base || cur->p_paddr >= ctx->memend)
				printf("Error: ssboot cmdline address is invalid (0x%08llx)\n",
						cur->p_paddr);
			ctx->cmdline_paddr = (void *)cur->p_paddr;
			ctx->cmdline_maxlen = cur->p_memsz;
		}
#endif
		if (cur->p_type != PT_LOAD)
			continue;

		if (i > 0 && prev->p_type == PT_LOAD && cur->p_paddr < prev->p_paddr) {
			printf("Error: elf segments not in sorted order! Aborting...\n");
			return -1;
		}

		memstart = filestart + cur->p_offset;
		if ((cur->p_paddr == memstart && cur->p_memsz > cur->p_filesz) ||
				cur->p_paddr > memstart) {
			printf("Error: elf segment will overwrite itself! Aborting...\n");
			return -1;
		}

		memcpy((void *)cur->p_paddr, (void *)(filestart + cur->p_offset),
				cur->p_filesz);
		if (cur->p_filesz != cur->p_memsz)
			memset((void *)cur->p_paddr + cur->p_filesz, 0x00,
					cur->p_memsz - cur->p_filesz);
		prev = cur;
	}
#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("\tRestore time: %lu ms\n", time);
#endif
	return 0;

}

/* Decompress an LZ4-compressed snapshot image in memory into an ELF file im memory */
static int load_lz4_from_mem(struct ss_ctx *ctx, void *addr)
{
	int ret;
	u32 magic;
	u8 flags, block_desc;
	u8 version, independent_blocks;
	u8 max_block_size;
	bool has_content_size, has_content_checksum, has_block_checksum;
	loff_t lz4_file_size, lz4_data_size, elf_file_size, dec_size;
	ulong memstart_lz4, memstart_elf;
	const int lz4_block_sizes[] = {SZ_64K, SZ_256K, SZ_1M, SZ_4M};
	void *in;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
	printf("## Decompressing LZ4 snapshot from address 0x%08lx...\n", (ulong)addr);

	in		= addr;
	magic		= get_unaligned_le32(in);
	in		+= sizeof(u32);
	flags		= *(u8 *)in;
	in		+= sizeof(u8);
	block_desc	= *(u8 *)in;
	in		+= sizeof(u8);

	version			= LZ4F_VERSION(flags);
	independent_blocks	= LZ4F_BLOCK_INDEP(flags);
	has_block_checksum	= LZ4F_HAS_BLOCK_CKSUM(flags);
	has_content_size	= LZ4F_HAS_CONTENT_SIZE(flags);
	has_content_checksum	= LZ4F_HAS_CONTENT_CKSUM(flags);
	max_block_size		= LZ4F_MAX_BLOCK_SIZE(block_desc);

	if (magic != LZ4F_MAGIC || version != 1 || independent_blocks != 1)
		return -EPROTONOSUPPORT;
	if ((flags & LZ4F_FLAGS_RESERVED_MASK) ||
			(block_desc & LZ4F_BLOCK_DESC_RESERVED_MASK))
		return -EINVAL;

	ctx->has_block_checksum	  = has_block_checksum;
	ctx->has_content_checksum = has_content_checksum;

	if (has_content_size) {
		elf_file_size = *(u64 *)in;
		in += sizeof(u64);
	} else {
		printf("Error: content size not available!\n");
		return -1;
	}
	in += sizeof(u8);

	if (max_block_size < 4 || max_block_size > 7)
		return -EINVAL;
	ctx->max_block_size = lz4_block_sizes[max_block_size - 4];

	lz4_data_size = get_lz4_file_size(in, ctx);
	if (lz4_data_size == 0) {
		printf("Error: cannot determine lz4 file size!\n");
		return -1;
	}
	lz4_file_size = in - addr + lz4_data_size;

	if (ctx->total_read_len && lz4_file_size != ctx->total_read_len) {
		printf("Error: lz4 file truncated!\n");
		return -1;
	} else {
		ctx->total_read_len = 0;
	}
	ctx->image_size_lz4 = lz4_file_size;

	printf("\tLZ4 snapshot image size: %llu.%03llu MB\n",
			ctx->image_size_lz4 >> 20, FLOAT(ctx->image_size_lz4, 10));

	if (have_enough_memory(ctx, elf_file_size) < 0)
		return -1;

	memstart_elf = ctx->memend - elf_file_size;
	memstart_lz4 = ctx->memend + LMB_BUFFER_SIZE - lz4_data_size;

	/*
	 * if we have the file loaded after the intended memstart, we are good!
	 * if not, we need to shift it forward to enable in-place loading.
	 * note that the uboot fs load commands fail if the load results in
	 * an lmb region overlap, so we are good either way!
	 */
	if ((ulong)in < memstart_lz4)
		memmove((void *)memstart_lz4, in, lz4_data_size);
	else
		memstart_lz4 = (ulong)in;

	/* allocate a temporary buffer for storing an LZ4 block for decomression */
	ctx->tbuf.size = ctx->max_block_size;
	ctx->tbuf.data = malloc(ctx->tbuf.size);
	if (!ctx->tbuf.data) {
		printf("Error: cannot allocate %llu B memory for decompression buffer!\n",
				ctx->tbuf.size);
		return -ENOMEM;
	}

	/*
	 * a buffer on heap is not needed for storing the LZ4 or ELF file since most of the
	 * u-boot memory is unused at this time and can be used as a buffer.
	 */
	ctx->cbuf.size = lz4_data_size;
	ctx->cbuf.data = (void *)memstart_lz4;
	ctx->cbuf.start = 0;

	ctx->ebuf.size = elf_file_size;
	ctx->ebuf.data = (void *)memstart_elf;
	ctx->ebuf.start = 0;

	dec_size = 0;
	do {
		ret = decompress_lz4(1, ctx);
		if (ret < 0) {
			printf("Error: lz4 decompression failed (%d)!\n", ret);
			return -1;
		}
		dec_size += ret;
		memstart_lz4 = (ulong)buf_cur(&ctx->cbuf);
		if (memstart_elf + ctx->ebuf.start > memstart_lz4) {
			printf("Error: elf data overwrites the lz4 data in memory!\n");
			return -1;
		}
	} while (ret > 0);
	if (dec_size != elf_file_size) {
		printf("Error: content size wrong! %lld %lld\n", dec_size, elf_file_size);
		return -1;
	}

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("\tDecompression time: %lu ms\n", time);
#endif
	return load_elf_from_mem(ctx, (void *)memstart_elf);
}

/* Verify a snapshot image file embedded in a FIT image for secure boot */
static ulong load_from_fit_image(struct ss_ctx *ctx, void *addr, const char *uname, ulong *data, ulong *len)
{
	int		verify;
	const void	*fit_hdr;
	int		noffset;
	const void	*fit_data;
	size_t		fit_len;
	uint8_t		comp;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
	int ret;

	verify = env_get_yesno("verify");

	printf("## Verifying '%s' subimage from FIT image "
			"at 0x%08lx...\n", uname, (ulong)addr);

	fit_hdr = (const void *)addr;
	ret = fit_check_format(fit_hdr, IMAGE_SIZE_INVAL);
	/*
	 * Note to backporters: the return value from fit_check_format
	 * is 0 for current u-boot but in the past success returned 1!!
	 */
	if (ret) {
		printf("Error: bad FIT image format: %d\n", ret);
		return -1;
	}

	/* get subimage node offset */
	noffset = fit_image_get_node(fit_hdr, uname);
	if (noffset < 0) {
		printf("Error: can't find '%s' FIT subimage\n", uname);
		return -1;
	}

	/* verify integrity */
	if (verify) {
		if (!fit_image_verify(fit_hdr, noffset)) {
			printf("Error: bad data hash\n");
			return -1;
		}
	}

	/* get subimage/external data address and length */
	if (fit_image_get_data_and_size(fit_hdr, noffset,
				       &fit_data, &fit_len)) {
		puts("Error: could not find script subimage data\n");
		return -1;
	}

	if (fit_image_get_comp(fit_hdr, noffset, &comp)) {
		puts("Error: could not find script subimage compression type\n");
		return -1;
	}

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("\tFIT Image verification time: %lums\n", time);
#endif
	*data = (ulong)fit_data;
	*len = (ulong)fit_len;

	/*
	 * If the FIT was originally loaded as a file total_read_len will be
	 * the total FIT file length. Switch it for the snapshot image len.
	 */
	ctx->total_read_len = fit_len;

	return 0;
}

/* Load snapshot image data from a snapshot file on the disk */
static int load_snapshot_from_file(struct ss_ctx *ctx, void **addr)
{
	int ret;
	loff_t file_size, actread;
	ulong offset;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
	printf("## Loading snapshot file '%s' from disk...\n", ctx->filename);

	if (fs_set_blk_dev(ctx->interface, ctx->devpart, ctx->fs_type)) {
		printf("Error: setting partition %s on interface %s failed!\n",
				ctx->devpart, ctx->interface);
		return -1;
	}

	if (fs_size(ctx->filename, &file_size) < 0) {
		printf("Error: cannot determine the file size!\n");
		return -1;
	}

	if (have_enough_memory(ctx, file_size) < 0)
		return -1;

	offset = ALIGN(ctx->memend + LMB_BUFFER_SIZE - file_size, ARCH_DMA_MINALIGN);

	if (fs_set_blk_dev(ctx->interface, ctx->devpart, ctx->fs_type)) {
		printf("Error: setting partition %s on interface %s failed!\n",
				ctx->devpart, ctx->interface);
		return -1;
	}

	/* NOTE: directly calling read() skips the lmb check! */
	ret = fs_read(ctx->filename, offset, 0, file_size, &actread);

	if (ret < 0 || file_size != actread) {
		printf("Error: failed to load the snapshot file!\n");
		return -1;
	}
	*addr = (void *)offset;
	ctx->total_read_len = actread;

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("\t%llu.%03llu MB read from disk in %lu ms\n", actread >> 20,
			FLOAT(actread, 10), time);
#endif
	return 0;
}

static int
do_ssbootmem(struct cmd_tbl *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	void *addr;
	char *ep;
	struct ss_ctx ctx;
	Elf_Addr entry;
	Elf_Ehdr *ehdr;
	u32 lz4f_magic;
	ulong data, len;
	const char *image_uname;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time = get_timer(0);
#endif
#ifdef CONFIG_CMD_SSBOOT_CMDLINEINFO
	const char *ssboot_bootcmd_line = env_get("ssboot_cmdline");
#endif
	if (argc < 2 || argc > 4)
		return CMD_RET_USAGE;

	addr = NULL;
	len = 0;
	memset(&ctx, 0, sizeof(ctx));

	switch (argc) {
	case 4: /* Read a file from disk into memory */
		ctx.interface	= argv[1];
		ctx.devpart	= argv[2];
		ctx.filename	= argv[3];
		ctx.fs_type	= SSBOOT_FS_TYPE;
		ctx.total_read_len = 0;
		ret = load_snapshot_from_file(&ctx, &addr);
		if (ret < 0) {
			printf("Error: loading snapshot file from disk failed!\n");
			return CMD_RET_FAILURE;
		}
	/* -- falling-through -- */
	case 3: /* FIT image loaded in memory */
		if (!addr) {
			addr = (void *)(simple_strtoul(argv[1], &ep, 16));
			if (ep == argv[1] || *ep != '\0') {
				printf("Error: %08lx is not a valid address!\n",
						(ulong)addr);
				return CMD_RET_FAILURE;
			}
		}
		if (genimg_get_format((void *)addr) == IMAGE_FORMAT_FIT) {
			image_uname = ((argc != 3) ? "snapshot-0" : argv[2]);
			data = 0;
			ret = load_from_fit_image(&ctx, addr, image_uname, &data, &len);
			if (ret < 0) {
				printf("Error: loading from FIT image failed!\n");
				return CMD_RET_FAILURE;
			} else
				addr = (void *)data;
		} else if (argc == 3) {
			printf("Error: no valid FIT image found at address %08lx!\n",
				(ulong)addr);
			return CMD_RET_FAILURE;
		}
	/* -- falling-through -- */
	case 2: /* ELF/LZ4 snapshot loaded in memory */
		if (!addr) {
			addr = (void *)(simple_strtoul(argv[1], &ep, 16));
			if (ep == argv[1] || *ep != '\0') {
				printf("Error: %08lx is not a valid address!\n",
						(ulong)addr);
				return CMD_RET_FAILURE;
			}
		}
		ehdr = (Elf_Ehdr *)addr;
		lz4f_magic = get_unaligned_le32(addr);

		if (lz4f_magic == LZ4F_MAGIC) {
			ctx.mode = SSBOOT_MODE_LZ4;
			ctx.image_size_lz4 = len;
		} else if (IS_ELF(*ehdr) && ehdr->e_type == ET_EXEC) {
			ctx.mode = SSBOOT_MODE_ELF;
			ctx.image_size_elf = len;
		} else {
			printf("Error: no valid ELF or LZ4 file found at address %08lx!\n",
					(ulong)addr);
			return CMD_RET_FAILURE;
		}
		break;
	default:
		return CMD_RET_FAILURE;
	}

	switch (ctx.mode) {
	case SSBOOT_MODE_LZ4:
		ret = load_lz4_from_mem(&ctx, addr);
		break;
	case SSBOOT_MODE_ELF:
		ret = load_elf_from_mem(&ctx, addr);
		break;
	default:
		return CMD_RET_FAILURE;
	}
	if (ret < 0)
		goto err;

#ifdef CONFIG_CMD_SSBOOT_CKSUM
	if (ssboot_verify_checksum(&ctx.elf.ehdr, ctx.elf.phdrs) < 0) {
		printf("Error: checksum verification failed\n");
		goto err;
	}
#endif
	entry = ctx.elf.ehdr.e_entry;
	free(ctx.elf.phdrs);

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	printf("Total time taken by Snapshot Boot: %lums\n", time);
#endif
#ifdef CONFIG_CMD_SSBOOT_CMDLINEINFO
	if (ctx.cmdline_paddr)
		strlcpy(ctx.cmdline_paddr, ssboot_bootcmd_line, ctx.cmdline_maxlen);
	printf("### Booting snapshot image ###\n");
	printf("\tEntry Address: 0x%8p\n", (void *)entry);
	printf("\tCommand line : %s\n", (char *)ctx.cmdline_paddr);
#else
	printf("## Booting snapshot image at 0x%8p ...\n", (void *)entry);
#endif

	cleanup_before_linux();
	/*
	 * Please don't do anything after cleanup_before_linux() and before jumping to the boot
	 * entry address. All interrupts and caches are disabled at this point.
	 */
	((void (*)(void))entry)();

	return CMD_RET_SUCCESS;

err:
	free(ctx.elf.phdrs);
	return CMD_RET_FAILURE;
}

U_BOOT_CMD(
	ssbootmem,      4,      0,      do_ssbootmem,
	"Load an ssboot snapshot image into memory and boot into Linux\n",
	"\tsupported image formats: ELF, LZ4, FIT\n"
	"<interface> <dev:part> <filename>\n"
	"\t- Load image 'filename' from 'dev:part' on 'interface'\n"
	"<addr> [FIT subimage]\n"
	"\t- Load image from memory address 'addr'\n"
);
