// SPDX-License-Identifier: GPL-2.0
/*
 *  Copyright 2022 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <elf.h>
#include <linux/ctype.h>
#include <linux/bitfield.h>

/*
 * The required malloc memory for various u-boot subsystems such as filesystems,
 * drivers, and for other commands.
 */
#define MALLOC_RESERVED_SIZE	SZ_4M
/*
 * Needed for allocating various temporary buffers in ssbootmem command
 */
#define SSBOOTMEM_BUFFER_SIZE	SZ_8M

#if (CONFIG_SYS_MALLOC_LEN < MALLOC_RESERVED_SIZE + SSBOOTMEM_BUFFER_SIZE)
#error "ssboot: increase the CONFIG_SYS_MALLOC_LEN value in the device defconfig file"
#endif

/*
 * During lmb reserve, extra 4KB space is reserved below the sp value for arm.
 * This enables us to spill over some of the in-memory snapshot file data over
 * to the reserved region. We make use of some of this "buffer" to avoid extra
 * memmove operations.
 * Ref commit: fcfa696b3a354ab1e16601683c61f671487aded7,
 * This value was later increased to 16K.
 * Ref commit: b1d0487cfe8dfaa5c4124bbe504b120b705f2d8e
 */
#define LMB_BUFFER_SIZE SZ_2K

#ifdef CONFIG_CMD_SSBOOT_FS_TYPE
#define SSBOOT_FS_TYPE	CONFIG_CMD_SSBOOT_FS_TYPE
#else
#define SSBOOT_FS_TYPE	FS_TYPE_ANY
#endif

#ifdef CONFIG_CMD_SSBOOT_CKSUM

#define elf_cksum_algorithm(cksum_phdr)		((cksum_phdr)->p_offset)
#define elf_ehdr_cksum(cksum_phdr)		((cksum_phdr)->p_vaddr)
#define elf_phdr_cksum(cksum_phdr)		((cksum_phdr)->p_paddr)
#define elf_data_cksum(cksum_phdr)		((cksum_phdr)->p_filesz)

#define CKSUM_TYPE_CRC32	1

enum {
	SSBOOT_WITH_CRC32,
	SSBOOT_WITHOUT_CKSUM,
};
#endif

#define FLOAT(size, n) (((((size) >> (n)) & 0x3FF) * 1000) / 1024)

static u16 LZ4_readLE16(const void *src)
{
	return get_unaligned_le16(src);
}
static void LZ4_copy4(void *dst, const void *src)
{
	put_unaligned(get_unaligned((const u32 *)src), (u32 *)dst);
}
static void LZ4_copy8(void *dst, const void *src)
{
	put_unaligned(get_unaligned((const u64 *)src), (u64 *)dst);
}

typedef  uint8_t BYTE;
typedef uint16_t U16;
typedef uint32_t U32;
typedef  int32_t S32;
typedef uint64_t U64;

#define FORCE_INLINE static inline __attribute__((always_inline))

#include "../lib/lz4.c"

#define LZ4F_BLOCKUNCOMPRESSED_FLAG 0x80000000U

struct ss_buf {
	void *data;
	u64 start, size;
};
#define buf_len(buf) ((buf)->size - (buf)->start)
#define buf_cur(buf) ((buf)->data + (buf)->start)

struct ss_elf {
	Elf_Ehdr ehdr;
	Elf_Phdr *phdrs;
};

struct ss_ctx {
	int		mode;

	struct ss_elf	elf;
	struct ss_buf	ebuf;
	loff_t          image_size_elf;

	struct ss_buf	cbuf;
	loff_t          image_size_lz4;
	bool		has_block_checksum;
	bool		has_content_checksum;
	int		max_block_size;

	struct ss_buf	tbuf;

	int		fs_type;
	const char	*interface;
	const char	*devpart;
	const char	*filename;
	loff_t		total_read_len;

	ulong		memend;
	ulong		free_size;
#ifdef CONFIG_CMD_SSBOOT_CMDLINEINFO
	void		*cmdline_paddr;
	unsigned int	cmdline_maxlen;
#endif
};

enum {
	SSBOOT_MODE_LZ4,
	SSBOOT_MODE_ELF
};

/*
 * LZ4 compression
 *
 * Flags byte:
 * +---------+--------------+----------+---------+----------+----------+---------+
 * |   7-6   |       5      |     4    |    3    |     2    |     1    |    0    |
 * +---------+--------------+----------+---------+----------+----------+---------+
 * | Version |     Block    |   Block  | Content |  Content | Reserved | Dict ID |
 * |         | Independence | Checksum |   Size  | Checksum |          |         |
 * +---------+--------------+----------+---------+----------+----------+---------+
 *
 * Block descriptor byte:
 * +----------+----------+----------+
 * |     7    |   6-5-4  |  3-2-1-0 |
 * +----------+----------+----------+
 * | Reserved |   Block  | Reserved |
 * |          | Max Size |          |
 * +----------+----------+----------+
 */
#define LZ4F_VERSION(x)			(FIELD_GET(GENMASK(7, 6), (x)))
#define LZ4F_BLOCK_INDEP(x)		(FIELD_GET(BIT(5), (x)))
#define LZ4F_HAS_BLOCK_CKSUM(x)		(FIELD_GET(BIT(4), (x)))
#define LZ4F_HAS_CONTENT_SIZE(x)	(FIELD_GET(BIT(3), (x)))
#define LZ4F_HAS_CONTENT_CKSUM(x)	(FIELD_GET(BIT(2), (x)))
#define LZ4F_MAX_BLOCK_SIZE(x)		(FIELD_GET(GENMASK(6, 4), (x)))

#define LZ4F_FLAGS_RESERVED_MASK	0x03
#define LZ4F_BLOCK_DESC_RESERVED_MASK	0x8F
