// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright (c) 2016 Google, Inc
 * Written by Simon Glass <sjg@chromium.org>
 */
#ifndef DEBUG
#define DEBUG 1
#endif

#include <common.h>
#include <backlight.h>
#include <dm.h>
#include <dm/device.h> // ZZZ
#include <dm/device_compat.h> // ZZZ
#include <panel.h>
#include <asm/gpio.h>
#include <linux/err.h>
#include <power/regulator.h>
#include <mipi_dsi.h>

struct simple_panel_dsi_priv {
	struct udevice *reg;
/*	struct udevice *backlight;
	struct gpio_desc enable;*/
	unsigned int lanes;
	enum mipi_dsi_pixel_format format;
	unsigned long mode_flags;
};

static const struct display_timing default_timing = {
	.pixelclock.typ		= 148500000,
	.hactive.typ		= 1920,
	.hfront_porch.typ	= 88,
	.hback_porch.typ	= 148,
	.hsync_len.typ		= 44,
	.vactive.typ		= 1080,
	.vfront_porch.typ	= 4,
	.vback_porch.typ	= 36,
	.vsync_len.typ		= 5,
};

static int simple_panel_dsi_enable_backlight(struct udevice *dev)
{
	struct simple_panel_dsi_priv *priv __attribute__((unused)) = dev_get_priv(dev); /// ZZZ
	int ret;

	debug("simple_panel_dsi_enable_backlight()\n");

	{
		struct mipi_dsi_panel_plat *plat = dev_get_plat(dev); // ZZZ
		struct mipi_dsi_device *device = plat->device;

		ret = mipi_dsi_attach(device);
		if (ret < 0){
			debug("simple_panel_dsi_enable_backlight() mipi_dsi_attach Failed:%d\n", ret);
			return ret;	
		}
	}

/*	debug("%s: start, backlight = '%s'\n", __func__, priv->backlight->name);
	dm_gpio_set_value(&priv->enable, 1);
	ret = backlight_enable(priv->backlight);
	debug("%s: done, ret = %d\n", __func__, ret);
	if (ret)
		return ret;
*/
	return 0;
}

static int simple_panel_dsi_set_backlight(struct udevice *dev, int percent)
{
	struct simple_panel_dsi_priv *priv __attribute__((unused)) = dev_get_priv(dev); /// ZZZ
	int ret __attribute__((unused)); // ZZZ

	debug("simple_paneldsi_set_backlight()\n");


/*
	debug("%s: start, backlight = '%s'\n", __func__, priv->backlight->name);
	dm_gpio_set_value(&priv->enable, 1);
	ret = backlight_set_brightness(priv->backlight, percent);
	debug("%s: done, ret = %d\n", __func__, ret);
	if (ret)
		return ret;
*/
	return 0;
}

static int simple_panel_get_display_timing(struct udevice *dev,
					    struct display_timing *timings)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dev); // ZZZ
	struct mipi_dsi_device *device = plat->device;
	struct simple_panel_dsi_priv *priv = dev_get_priv(dev);

	debug("simple_panel_get_display_timing()\n");

	memcpy(timings, &default_timing, sizeof(*timings));

	/* fill characteristics of DSI data link */
	if (device) {
		device->lanes = priv->lanes;
		device->format = priv->format;
		device->mode_flags = priv->mode_flags;
	}

	return 0;
}


static int simple_panel_dsi_ofdata_to_platdata(struct udevice *dev)
{
	struct simple_panel_dsi_priv *priv = dev_get_priv(dev);
	int ret;

	debug("simple_panel_dsi_ofdata_to_platdata()\n");

	if (IS_ENABLED(CONFIG_DM_REGULATOR)) {
		ret = uclass_get_device_by_phandle(UCLASS_REGULATOR, dev,
						   "power-supply", &priv->reg);
		if (ret) {
			debug("%s: Warning: cannot get power supply: ret=%d\n",
			      __func__, ret);
			if (ret != -ENOENT)
				return ret;
		}
	}
/*
	ret = uclass_get_device_by_phandle(UCLASS_PANEL_BACKLIGHT, dev,
					   "backlight", &priv->backlight);
	if (ret) {
		debug("%s: Cannot get backlight: ret=%d\n", __func__, ret);
		return log_ret(ret);
	}
	ret = gpio_request_by_name(dev, "enable-gpios", 0, &priv->enable,
				   GPIOD_IS_OUT);
	if (ret) {
		debug("%s: Warning: cannot get enable GPIO: ret=%d\n",
		      __func__, ret);
		if (ret != -ENOENT)
			return log_ret(ret);
	}
*/
	return 0;
}

static int simple_panel_dsi_probe(struct udevice *dev)
{
	struct simple_panel_dsi_priv *priv = dev_get_priv(dev);
	int ret;

	debug("simple_panel_dsi_probe()\n");

	priv->format = MIPI_DSI_FMT_RGB888;
	priv->mode_flags = MIPI_DSI_MODE_VIDEO | MIPI_DSI_MODE_VIDEO_SYNC_PULSE |
			  MIPI_DSI_MODE_EOT_PACKET | MIPI_DSI_MODE_VIDEO_HSE;

	if (IS_ENABLED(CONFIG_DM_REGULATOR) && priv->reg) {
		debug("%s: Enable regulator '%s'\n", __func__, priv->reg->name);
		ret = regulator_set_enable(priv->reg, true);
		if (ret)
			return ret;
	}

	ret = dev_read_u32(dev, "dsi-lanes", &priv->lanes);
	if (ret) {
		dev_err(dev, "Failed to get dsi-lanes property (%d)\n", ret);
		return ret;
	}

	if (priv->lanes < 1 || priv->lanes > 4) {
		dev_err(dev, "Invalid dsi-lanes: %d\n", priv->lanes);
		return -EINVAL;
	}

	return 0;
}

static const struct panel_ops simple_panel_dsi_ops = {
	.enable_backlight	= simple_panel_dsi_enable_backlight,
	.get_display_timing = simple_panel_get_display_timing,
	.set_backlight		= simple_panel_dsi_set_backlight,
};

static const struct udevice_id simple_panel_dsi_ids[] = {
	{ .compatible = "simple-panel-dsi" },
	{ }
};

U_BOOT_DRIVER(simple_panel_dsi) = {
	.name	= "simple_panel_dsi",
	.id	= UCLASS_PANEL,
	.of_match = simple_panel_dsi_ids,
	.ops	= &simple_panel_dsi_ops,
	.of_to_plat	= simple_panel_dsi_ofdata_to_platdata, // ZZZ
	.probe		= simple_panel_dsi_probe,
	.priv_auto	= sizeof(struct simple_panel_dsi_priv), // ZZZ
};
